"""
Test Cases
* `list` from an empty database
* `list` from a non-empty database
"""
import pytest
from cards import Card

def test_list_no_cards(cards_db):
    """Empty db, empty list"""
    assert cards_db.list_cards() == []


def test_list_several_cards(cards_db):
    """"
    Given a variety of cards, make sure they get returned.
    """
    orig = [
        Card("foo"),
        Card("bar", owner="me"),
        Card("baz", owner="you", state="in prog"),
    ]

    for c in orig:
        cards_db.add_card(c)

    the_list = cards_db.list_cards()

    assert len(the_list) == len(orig)
    for c in orig:
        assert c in the_list

# list filter
# - no owner
# - specific owner
# - specific state
# - owner and state

@pytest.fixture()
def known_set():
    return [Card(summary="zero", owner="Brian", state="todo"),
            Card(summary="one", owner="Brian", state="in prog"),
            Card(summary="two", owner="Brian", state="done"),

            Card(summary="three", owner="Okken", state="todo"),
            Card(summary="four", owner="Okken", state="in prog"),
            Card(summary="five", owner="Okken", state="done"),

            Card(summary="six", state="todo"),
            Card(summary="seven", state="in prog"),
            Card(summary="eight", state="done")]

@pytest.fixture()
def db_filled(cards_db, known_set):
    for c in known_set:
        cards_db.add_card(c)
    return cards_db

@pytest.mark.parametrize('owner_, state_, expected_indices', [
    ("", None, (6, 7, 8)),
    ("Brian", None, (0, 1, 2)),
    ("Okken", None, (3, 4, 5)),
    (None, "todo", (0, 3, 6)),
    (None, "in prog", (1, 4, 7)),
    (None, "done", (2, 5, 8)),
    ("Brian", "todo", (0,)),
], ids=str)
def test_list_filter(db_filled, known_set, owner_, state_, expected_indices):
    result = db_filled.list_cards(owner=owner_, state=state_)
    assert len(result) == len(expected_indices)
    for i in expected_indices:
        assert known_set[i] in result




import pytest
import cards
from cards import Card


@pytest.fixture(scope="session")
def db_path(tmp_path_factory):
    """Path to temporary database"""
    return tmp_path_factory.mktemp("cards_db")


@pytest.fixture(scope="session")
def session_cards_db(db_path):
    """CardsDB"""
    db_ = cards.CardsDB(db_path)
    yield db_
    db_.close()


@pytest.fixture(scope="function")
def cards_db(session_cards_db, request, faker):
    db = session_cards_db
    db.delete_all()
    # support for `@pytest.mark.num_cards(<some number>)`
    faker.seed_instance(101) # random seed
    m = request.node.get_closest_marker('num_cards')
    if m and len(m.args) > 0:
        num_cards = m.args[0]
        for _ in range(num_cards):
            db.add_card(Card(summary=faker.sentence(),
                             owner=faker.first_name()))
    return db







"""
Test Cases
* `version` returns the correct version
"""
import re
import cards


def test_version():
    """
    There is no api for version other than cards.__version__.
    However, we do expect it to be:
    - a string containing a version in the form of "x.y.z"
    So, why not, let's test for that.
    """
    version = cards.__version__
    assert re.match(r"\d+\.\d+\.\d+", version)



from typer.testing import CliRunner
import cards
import pytest
import shlex

runner = CliRunner()

@pytest.fixture()
def cards_cli_no_redirect():
    def run_cli(command_string):
        command_list = shlex.split(command_string)
        result = runner.invoke(cards.cli.app, command_list)
        output = result.stdout.rstrip()
        return output
    return run_cli

@pytest.fixture()
def cards_cli(cards_cli_no_redirect, db_path, monkeypatch, cards_db):
    monkeypatch.setenv("CARDS_DB_DIR", db_path.as_posix())
    return cards_cli_no_redirect



"""Command Line Interface (CLI) for cards project."""
import os
from io import StringIO
import pathlib
import rich
from rich.table import Table
from contextlib import contextmanager
from typing import List
import cards
import typer


app = typer.Typer(add_completion=False)


@app.command()
def version():
    """Return version of cards application"""
    print(cards.__version__)


@app.command()
def add(
    summary: List[str], owner: str = typer.Option(None, "-o", "--owner")
):
    """Add a card to db."""
    summary = " ".join(summary) if summary else None
    with cards_db() as db:
        db.add_card(cards.Card(summary, owner, state="todo"))


@app.command()
def delete(card_id: int):
    """Remove card in db with given id."""
    with cards_db() as db:
        try:
            db.delete_card(card_id)
        except cards.InvalidCardId:
            print(f"Error: Invalid card id {card_id}")


@app.command("list")
def list_cards(
    owner: str = typer.Option(None, "-o", "--owner"),
    state: str = typer.Option(None, "-s", "--state"),
):
    """
    List cards in db.
    """
    with cards_db() as db:
        the_cards = db.list_cards(owner=owner, state=state)
        table = Table(box=rich.box.SIMPLE)
        table.add_column("ID")
        table.add_column("state")
        table.add_column("owner")
        table.add_column("summary")
        for t in the_cards:
            owner = "" if t.owner is None else t.owner
            table.add_row(str(t.id), t.state, owner, t.summary)
        out = StringIO()
        rich.print(table, file=out)
        print(out.getvalue())


@app.command()
def update(
    card_id: int,
    owner: str = typer.Option(None, "-o", "--owner"),
    summary: List[str] = typer.Option(None, "-s", "--summary"),
):
    """Modify a card in db with given id with new info."""
    summary = " ".join(summary) if summary else None
    with cards_db() as db:
        try:
            db.update_card(
                card_id, cards.Card(summary, owner, state=None)
            )
        except cards.InvalidCardId:
            print(f"Error: Invalid card id {card_id}")


@app.command()
def start(card_id: int):
    """Set a card state to 'in prog'."""
    with cards_db() as db:
        try:
            db.start(card_id)
        except cards.InvalidCardId:
            print(f"Error: Invalid card id {card_id}")


@app.command()
def finish(card_id: int):
    """Set a card state to 'done'."""
    with cards_db() as db:
        try:
            db.finish(card_id)
        except cards.InvalidCardId:
            print(f"Error: Invalid card id {card_id}")


@app.command()
def config():
    """List the path to the Cards db."""
    with cards_db() as db:
        print(db.path())


@app.command()
def count():
    """Return number of cards in db."""
    with cards_db() as db:
        print(db.count())


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """
    Cards is a small command line task tracking application.
    """
    if ctx.invoked_subcommand is None:
        list_cards(owner=None, state=None)


def get_path():
    db_path_env = os.getenv("CARDS_DB_DIR", "")
    if db_path_env:
        db_path = pathlib.Path(db_path_env)
    else:
        db_path = pathlib.Path.home() / "cards_db"
    return db_path


@contextmanager
def cards_db():
    db_path = get_path()
    db = cards.CardsDB(db_path)
    yield db
    db.close()



import cards


expected_output = """\

  ID   state   owner   summary    
 ──────────────────────────────── 
  1    todo            some task  
  2    todo            another
"""

def test_list(cards_db, cards_cli):
    cards_db.add_card(cards.Card("some task"))
    cards_db.add_card(cards.Card("another"))
    output = cards_cli("list")
    assert output.strip() == expected_output.strip()

def test_main(cards_db, cards_cli):
    cards_db.add_card(cards.Card("some task"))
    cards_db.add_card(cards.Card("another"))
    output = cards_cli("")
    assert output.strip() == expected_output.strip()




import pytest


@pytest.mark.parametrize("command", ["delete 25",
                                     "start 25",
                                     "finish 25",
                                     "update 25 -s foo -o brian"])
def test_invalid_card_id(cards_db, command, cards_cli):
    out = cards_cli(command)
    assert out == "Error: Invalid card id 25"


def test_missing_summary(cards_db, cards_cli):
    out = cards_cli("add")
    assert "Error: Missing argument 'SUMMARY...'" in out



import cards

def test_update(cards_db, cards_cli):
    i = cards_db.add_card(cards.Card("foo"))
    cards_cli(f"update {i} -o okken -s something")
    expected = cards.Card("something", owner="okken", state="todo")
    actual = cards_db.get_card(i)
    assert actual == expected


"""
Test Cases
* `start` from "todo", "in prog", and "done" states
* `start` an invalid id
"""
import pytest
from cards import Card, InvalidCardId


@pytest.mark.parametrize("start_state", ("todo", "in prog", "done"))
def test_start(cards_db, start_state):
    """
    End state should be "in prog"
    """
    c = Card("foo", state=start_state)
    i = cards_db.add_card(c)
    cards_db.start(i)
    c = cards_db.get_card(i)
    assert c.state == "in prog"


def test_start_non_existent(cards_db):
    """
    Shouldn't be able to start a non-existent card.
    """
    i = 123  # any number will do, db is empty
    with pytest.raises(InvalidCardId):
        cards_db.start(i)



"""
Test Cases:
* `delete` one from a database with more than one
* `delete` the last card
* `delete` a non-existent card
"""
import pytest
from cards import Card, InvalidCardId

@pytest.fixture()
def three_cards(cards_db):
    i = cards_db.add_card(Card("foo"))
    j = cards_db.add_card(Card("bar"))
    k = cards_db.add_card(Card("baz"))
    return (i, j, k) # ids for the cards

def test_delete_from_many(cards_db, three_cards):
    """
    Count should go from 3 to 2
    And card shouldn't be retrievable.
    But the rest should be.
    """
    (i, j, k) = three_cards # ids
    id_to_delete = j
    ids_still_there = (i, k)

    cards_db.delete_card(id_to_delete)

    assert cards_db.count() == 2
    # card should not be retrievable after deletion
    with pytest.raises(InvalidCardId):
        cards_db.get_card(id_to_delete)
    # non-deleted cards should still be retrievable
    for i in ids_still_there:
        # just making sure this doesn't throw an exception
        cards_db.get_card(i)


def test_delete_last_card(cards_db):
    """
    Count should be back to 0
    And card shouldn't be retrievable.
    """
    i = cards_db.add_card(Card("foo"))
    cards_db.delete_card(i)
    assert cards_db.count() == 0
    with pytest.raises(InvalidCardId):
        cards_db.get_card(i)


def test_delete_non_existent(cards_db):
    """
    Shouldn't be able to start a non-existent card.
    """
    i = 123  # any number will do, db is empty
    with pytest.raises(InvalidCardId):
        cards_db.delete_card(i)



"""
Test Cases
* `update` the owner of a card
* `update` the summary of a card
* `update` owner and summary of a card at the same time
* `update` a non-existent card
"""
import pytest
from cards import Card, InvalidCardId


def test_update_owner(cards_db):
    """
    summary and state should stay the same
    owner should change
    """
    i = cards_db.add_card(Card("foo", owner="me"))
    cards_db.update_card(i, Card(owner="not me", state=None))

    mod = cards_db.get_card(i)
    assert mod == Card("foo", owner="not me")


def test_update_summary(cards_db):
    """
    owner and state should stay the same
    summary should change
    """
    i = cards_db.add_card(Card("foo", owner="me", state="done"))
    cards_db.update_card(i, Card(summary="bar", state=None))

    mod = cards_db.get_card(i)
    assert mod == Card("bar", owner="me", state="done")


def test_update_both(cards_db):
    """
    state should stay the same
    owner and summary should change
    """
    i = cards_db.add_card(Card("foo", owner="me"))
    cards_db.update_card(i, Card(summary="bar", owner="not me"))

    mod = cards_db.get_card(i)
    assert mod == Card("bar", owner="not me", state="todo")


def test_update_non_existent(cards_db):
    """
    Shouldn't be able to update a non-existent card.
    """
    i = 123  # any number will do, db is empty
    with pytest.raises(InvalidCardId):
        cards_db.update_card(i, Card(summary="bar", owner="not me"))


"""
API for the cards project
"""
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field

from .db import DB

__all__ = [
    "Card",
    "CardsDB",
    "CardsException",
    "MissingSummary",
    "InvalidCardId",
]


@dataclass
class Card:
    summary: str = None
    owner: str = None
    state: str = "todo"
    id: int = field(default=None, compare=False)

    @classmethod
    def from_dict(cls, d):
        return Card(**d)

    def to_dict(self):
        return asdict(self)


class CardsException(Exception):
    pass


class MissingSummary(CardsException):
    pass


class InvalidCardId(CardsException):
    pass


class CardsDB:
    def __init__(self, db_path):
        self._db_path = db_path
        self._db = DB(db_path, ".cards_db")

    def add_card(self, card: Card):
        """Add a card, return the id of card."""
        if not card.summary:
            raise MissingSummary
        if card.owner is None:
            card.owner = ""
        card_id = self._db.create(card.to_dict())
        self._db.update(card_id, {"id": card_id})
        return card_id

    def get_card(self, card_id: int):
        """Return a card with a matching id."""
        db_item = self._db.read(card_id)
        if db_item is not None:
            return Card.from_dict(db_item)
        else:
            raise InvalidCardId(card_id)

    def list_cards(self, owner=None, state=None):
        """Return a list of cards."""
        all_cards = self._db.read_all()
        if (owner is not None) and (state is not None):
            return [
                Card.from_dict(t)
                for t in all_cards
                if (t["owner"] == owner and t["state"] == state)
            ]
        elif owner is not None:
            return [
                Card.from_dict(t) for t in all_cards if t["owner"] == owner
            ]
        elif state is not None:
            return [
                Card.from_dict(t) for t in all_cards if t["state"] == state
            ]
        else:
            return [Card.from_dict(t) for t in all_cards]

    def count(self):
        """Return the number of cards in db."""
        return self._db.count()

    def update_card(self, card_id: int, card_mods: Card):
        """Update a card with modifications."""
        try:
            self._db.update(card_id, card_mods.to_dict())
        except KeyError as exc:
            raise InvalidCardId(card_id) from exc

    def start(self, card_id: int):
        """Set a card state to 'in prog'."""
        self.update_card(card_id, Card(state="in prog"))

    def finish(self, card_id: int):
        """Set a card state to 'done'."""
        self.update_card(card_id, Card(state="done"))

    def delete_card(self, card_id: int):
        """Remove a card from db with given card_id."""
        try:
            self._db.delete(card_id)
        except KeyError as exc:
            raise InvalidCardId(card_id) from exc

    def delete_all(self):
        """Remove all cards from db."""
        self._db.delete_all()

    def close(self):
        self._db.close()

    def path(self):
        return self._db_path



"""
Test Cases
* `add` to an empty database, with summary
* `add` to a non-empty database, with summary
* `add` a card with both summary and owner set
* `add` a card with a missing summary
* `add` a duplicate card
"""
import pytest
from cards import Card, MissingSummary

def test_add_from_empty(cards_db):
    """
    count should be 1 and card retrievable
    """
    c = Card(summary="do something")
    i = cards_db.add_card(c)
    assert cards_db.count() == 1
    assert cards_db.get_card(i) == c


@pytest.mark.num_cards(3)
def test_add_from_nonempty(cards_db):
    """
    count should increase by 1 and card retrievable
    """
    c = Card(summary="do something")
    i = cards_db.add_card(c)
    assert cards_db.count() == 4
    assert cards_db.get_card(i) == c

def test_add_with_summary_and_owner(cards_db):
    """
    count should be 1 and card retrievable
    """
    c = Card(summary="do something", owner="Brian")
    i = cards_db.add_card(c)
    assert cards_db.count() == 1
    assert cards_db.get_card(i) == c

def test_add_missing_summary(cards_db):
    """
    Should raise an exception
    """
    c = Card()
    with pytest.raises(MissingSummary):
        cards_db.add_card(c)

def test_add_duplicate(cards_db):
    """
    Duplicates allowed, both retrievable, separate indices
    """
    c = Card(summary="do something")
    i_1 = cards_db.add_card(c)
    i_2 = cards_db.add_card(c)
    c1 = cards_db.get_card(i_1)
    c2 = cards_db.get_card(i_2)
    assert i_1 != i_2
    assert c1 == c2 == c



import cards

def test_add(cards_db, cards_cli):
    cards_cli("add some task")
    expected = cards.Card("some task", owner="", state="todo")
    all = cards_db.list_cards()
    assert len(all) == 1
    assert all[0] == expected


def test_add_with_owner(cards_db, cards_cli):
    cards_cli("add some task -o brian")
    expected = cards.Card("some task", owner="brian", state="todo")
    all = cards_db.list_cards()
    assert len(all) == 1
    assert all[0] == expected


"""
DB for the cards project
"""


import tinydb


class DB:
    def __init__(self, db_path, db_file_prefix):
        self._db = tinydb.TinyDB(
            db_path / f"{db_file_prefix}.json", create_dirs=True
        )

    def create(self, item: dict) -> int:
        id = self._db.insert(item)
        return id

    def read(self, id: int):
        item = self._db.get(doc_id=id)
        return item

    def read_all(self):
        return self._db

    def update(self, id: int, mods) -> None:
        changes = {k: v for k, v in mods.items() if v is not None}
        self._db.update(changes, doc_ids=[id])

    def delete(self, id: int) -> None:
        self._db.remove(doc_ids=[id])

    def delete_all(self) -> None:
        self._db.truncate()

    def count(self) -> int:
        return len(self._db)

    def close(self):
        self._db.close()



def test_config(cards_cli, db_path):
    assert cards_cli("config") == str(db_path)

def test_config_normal_path(db_path, cards_cli_no_redirect):
    cards_cli = cards_cli_no_redirect
    assert cards_cli("config") != str(db_path)



"""
Test Cases
* `count` from an empty database
* `count` with one item
* `count` with more than one item
"""
import pytest


def test_count_no_cards(cards_db):
    assert cards_db.count() == 0

@pytest.mark.num_cards(1)
def test_count_one_card(cards_db):
    assert cards_db.count() == 1

@pytest.mark.num_cards(3)
def test_count_three_cards(cards_db):
    assert cards_db.count() == 3
